/*
  Example of compile time ordering
*/

#include <iostream>
#include <fstream>
#include <cmath>

#include <algorithm>

// standard bubble sort
void
bubbleSort( int data[], int N ) {
  for ( int i = N-1 ; i > 0 ; --i )
    for ( int j = 0 ; j < i ; ++j )
      if ( data[j] > data[i] )
        std::swap( data[i], data[j] ) ;
}

// recursive ordering of a vector
void
recursiveSort( int data[], int N ) {
  // loop on the vector and put maximum at the end of the vector
  for ( int j = 0 ; j < N-1 ; ++j )
    if ( data[j] > data[j+1] )
      std::swap( data[j], data[j++] ) ;
  // at the end of this loop
  // data[N-1] is the maximum of the vector
  // if N > 2 reapply the algorithm to the vector vec[0..N-2]
  if ( N > 2 ) recursiveSort( data, N-1 ) ;
}

// recursive ordering of a vector using template metaprogram


/*
  This class implement method compareAndSwap which 
  compare data[I] and data[J] and if necessary swap it.
 */
template <int I, int J>
class Swap {
public:
  static inline void
  compareAndSwap( int data[] ) {
    if ( data[I] > data[J] ) std::swap( data[I], data[J] ) ;
  }
} ;

/*
  this class implement method loop which 
  recursively get the maximum of the vector data[0..N-1]
  and put in data[N-1]
 */
template <int N>
class MaxLoop {
public:
  static inline void
  loop( int data[] ) {
    MaxLoop<N-1>::loop(data) ;
    Swap<N-2,N-1>::compareAndSwap(data) ;
  }
} ;

/*
  if the length of the vector is 1 nothing to do
 */
template <>
class MaxLoop<1> {
public:
  static inline void
  loop( int data[] ) {}
} ;

/*
  Sort the vector 
 */
template <int N>
class Sort {
public:
  static inline void
  doSort( int data[] ) {
    MaxLoop<N>::loop(data) ;
    Sort<N-1>::doSort(data) ;
  }
} ;

/*
  if the length of the vector is 1 nothing to do
 */
template <>
class Sort<1> {
public:
  static inline void
  doSort( int data[] ) {}
} ;

/*
   Check the ordering
 */

int
main() {

  {
    int data[]  = { 4, 3, -1, 2, 3, -2, 1, 4 } ;
    int const N = sizeof(data)/sizeof(data[0]) ;  
    bubbleSort( data, N ) ;
    for ( int i = 0 ; i < N ; ++i )
      std::cout << data[i] << ", " ;
    std::cout << "\n" ;
  }

  {
    int data[]  = { 4, 3, -1, 2, 3, -2, 1, 4 } ;
    int const N = sizeof(data)/sizeof(data[0]) ;  
    recursiveSort( data, N) ;
    for ( int i = 0 ; i < N ; ++i )
      std::cout << data[i] << ", " ;
    std::cout << "\n" ;
  }

  {
    int data[]  = { 4, 3, -1, 2, 3, -2, 1, 4 } ;
    int const N = sizeof(data)/sizeof(data[0]) ;  
    Sort<N>::doSort( data ) ;
    for ( int i = 0 ; i < N ; ++i )
      std::cout << data[i] << ", " ;
    std::cout << "\n" ;
  }
  
  return 0 ;
}
