/*
  Example of class used as function
*/

#include <iostream>
#include <fstream>
#include <cmath>

#include <algorithm>

/* 
  define the class stupid
  with the operator () defined and used as a function
*/

class stupid {
  double a ;
public:
  stupid( double _a ) : a(_a) {} 

  void setup( double x ) { a = x ; }
  
  double 
  operator () ( double x ) const
  { return sin(a*x) ; }
} ;

/*
  non template version:
  double
  midpoint( double  const a,
            double  const b,
            int const N,
            double (FUN*)(double) ) {
  }
  
  the function midpoint implement the numerical integration rule
  by midpoint formula.
  
*/

template <typename T>
T
midpoint( T   const a,
          T   const b,
          int const N,
          T (*FUN)(T) ) {
  T res = 0 ;
  T h   = (b-a)/N ;
  for ( int i = 1 ; i < N ; ++i )
    res += FUN( a + h * (i-0.5) ) ;
  return res * h ;
}

/*
  This version of midpoint integration rule is paramtetrized also 
  with the function argument.
  In this way FUN can be any thing that understand operator ()
  i.e. a function or an object with deefined the operator ()!.
*/

template <typename T, typename FTYPE>
T
midpoint2( T     const a,
           T     const b,
           int   const N,
           FTYPE FUN ) {
  T res = 0 ;
  T h   = (b-a)/N ;
  for ( int i = 1 ; i < N ; ++i )
    res += FUN( a + h * (i-0.5) ) ;
  return res * h ;
}

/*
  a simple function for test
*/
inline
double
fun( double a ) 
{ return a*a ; }

int
main() {
  stupid like_a_function(2) ;
  
  std::cout << like_a_function(1.23) << "\n" ;
  
  double res = midpoint2<double>( 0, 2*M_PI, 1000, like_a_function ) ;
  std::cout << "integral = " << res << "\n" ;

  res = midpoint2<double>( 0, 2*M_PI, 1000, fun ) ;
  std::cout << "integral = " << res << "\n" ;

  return 0 ;
}
